# ncsa.pm was patched to support wn-server by M. Feenstra on 20/09/2001
package Parser;

use strict;
use warnings;
use vars qw(%datafields %extra);

use Exporter;
use WebserverTools qw(%month_numbers detect_referer detect_browser detect_os
	extract_data);

@Parser::EXPORT_OK = qw(parse extra %datafields);

@Parser::ISA = ('Exporter');

$::CONTRIB{'Mark Feenstra'} = 'mark@itmon.nl';

# FIXME pass date and hour in seconds to properly format and sort.

$extra{'total'} = 0;
$extra{'days'} = {};
#$extra{'codered'} = 0;
#$extra{'nimda'} = 0;

%datafields = (
    host	=> 'key:host',
    user	=> 'key',
    date	=> 'key',
    hour	=> 'key',
    browser	=> 'key',
    os		=> 'key',
    domain	=> 'key:host',
    tld		=> 'key',
    file	=> 'key',
    referer	=> 'key',
    status	=> 'key',
    virtualhost	=> 'key',
    bytes	=> 'amount:bytes',
    requests	=> 'amount:number'
    );

# This needs to be done at runtime, since Data uses Setup which relies on
# %datafields to be defined  -octo
require Data;       
import Data qw#store#; 

print $/, __FILE__, q(: Using wnserver format) if $::DEBUG;    

return (1);

sub parse
{
	my $line = shift or return undef;
	if ($line =~ /^(\S+)\s(\S+)\s(.+?)\s\[(.+?)\]\s"(.+?)"\s(\d+)\s(\S+)\s<.*?><(.*?)>\s<(.*?)>\s<(.*?)>\s<.*?>(?:\s<(.*?)>)?$/)
#	if ($line =~ /^(\S+)\s(\S+)\s(\S+)\s\[([^\]]+)\]\s\"([^\"]+)\"\s(\d+)\s(\d+)\s\<[^\>]*\>\<([^\>]*)\>\s\<([^\>]*)\>\s\<([^\>]*)\>\s\<([^\>]*)\>$/)
	{
		my ($host, $ident, $user, $date, $request, $status,
			$bytes, $browser, $referer, $cookie, $nickname) =
		($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11);

		my ($day, $month, $year, $hour, $minute, $second) =
			$date =~ m#(\d\d)/(\w{3})/(\d{4}):(\d\d):(\d\d):(\d\d)#;

		my $n_date = $year . '-' . $month_numbers{$month} . '-' . $day;

		my ($method, $file, $params) =
			$request =~ m#(\S+) ([^\?]+)\??(\S*)#
				or return (0);
		
#		my $attack = 0;
#		if (lc ($file) eq '/default.ida') { ++$extra{'codered'}; $attack = 1; }
#		elsif (lc ($file) eq '/scripts/root.exe') { ++$extra{'nimda'}; $attack = 1; }
#		elsif ($file =~ /root\.exe|winnt/i) { $attack = 1; }
#		if ($attack and !$::config->{'inc_attacks'}) { return 'attack'; }

		if (($user ne '-') and ($status >= 400) and ($status < 500))
		{
			$user = 'invalid';
		}

		if ($user eq '-') { $user = 'unknown'; }
		if ($bytes eq '-') { $bytes = 0; }

		my ($domain, $tld);
		if ($host =~ m/([^\.]+)\.([a-z]{2,})$/i)
		{
			$tld = lc ($2);
			$domain = lc ($1) . ".$tld";
		}
		else
		{       
			$domain = 'UNRESOLVED';
			$tld = 'UNRESOLVED';
		}

		my $os = detect_os ($browser);
		my $browser_name = detect_browser ($browser);
		my @search_terms = extract_data ($referer);
		$referer = detect_referer ($referer);
		if ($referer eq '-') { $referer = ''; }
		
		if (!defined $extra{'days'}->{$n_date}) { $extra{'days'}->{$n_date} = 1; }
		++$extra{'total'};

		if (scalar @search_terms)
		{
			print $/, __FILE__, ": Search Terms: ",
				join (' ', @search_terms)
				if ($::DEBUG & 0x100);
			
			$extra{'search_terms'}->{$_}++ for (@search_terms);
		}
		
		my %combined = (
					'host'		=>	$host,
					'user'		=>	$user,
					'date'		=>	$n_date,
					'hour'		=>	$hour,
					'browser'	=>	$browser_name,
					'os'		=>	$os,
					'domain'	=>	$domain,
					'tld'		=>	$tld,
					'file'		=>	$file,
					'referer'	=>	$referer,
					'status'	=>	$status,
					'bytes'		=>	$bytes,
					'virtualhost'	=>	$nickname,
					'requests'	=>	1
										);
		store (\%combined);
	}
	elsif ($::DEBUG)
	{
		chomp ($line);
		print STDERR $/, __FILE__, ": Unable to parse: '$line'";
	}
}

sub extra
{
	my ($average, $days) = (0, 0);
	
	$days = scalar keys %{$extra{'days'}};
	return (0) unless ($days);
	
	$average = int (($extra{'total'} * 10) / $days) / 10;

	$::EXTRA->{'Total requests'} = $extra{'total'};
	$::EXTRA->{'Averange requests per day'} = $average;
	$::EXTRA->{'Reporting period'} = "$days days";
#	$::EXTRA->{'CodeRed attacks'} = $extra{'codered'};
#	$::EXTRA->{'Nimda attacks'} = $extra{'nimda'};
	
	my @sorted_terms = sort
		{ $extra{'search_terms'}->{$b} <=> $extra{'search_terms'}->{$a} }
		(keys %{$extra{'search_terms'}});
	
	if (@sorted_terms)
	{
		my $max = $extra{'search_terms'}->{$sorted_terms[0]};
		my @scalar_terms = ();
		
		while (@sorted_terms and
			($extra{'search_terms'}->{$sorted_terms[0]} / $max) > 0.1)
		{
			$_ = shift (@sorted_terms);
			
			push (@scalar_terms,
				sprintf ("%s (%u)",
					$_, $extra{'search_terms'}->{$_})
			);
		}
		$::EXTRA->{'Search terms used'} = join (', ', @scalar_terms);

		if (@sorted_terms)
		{
			my $skipped = scalar (@sorted_terms);
			$::EXTRA->{'Search terms used'} .= " [$skipped more skipped]";
		}
	}
}

1;
